package se.paxport.yggejonssoncrazysnake;

import se.citerus.crazysnake.Position;

import java.util.HashSet;
import java.util.Map;
import java.util.Set;

public class FruitDistance {

    private static final int TURNS_PER_GAME = 1000;

    private static final int MAX_CLOSE_EXTRA_FRUITS = 6;

    private static final int MAX_TOTAL_FRUIT_DIST = 120;

    private static final int AVERAGE_TURNS_TO_NEXT_FRUIT = 30;

    private final Position fruitPosition;

    private final int myDist, oppDist, turn;

    private final int[] turnsToGetExtraFruits = new int[MAX_CLOSE_EXTRA_FRUITS + 1];

    public FruitDistance(int turn, Position fruitPosition, int myDist, Integer oppDist) {
        this.turn = turn;
        this.fruitPosition = fruitPosition;
        this.myDist = myDist;
        this.oppDist = oppDist == null ? Integer.MAX_VALUE : oppDist;
    }

    public Position getFruitPosition() {
        return fruitPosition;
    }

    public int getMyDist() {
        return myDist;
    }

    public int getOppDist() {
        return oppDist;
    }

    public int compareFruitDistances(FruitDistance other) {
        int fruits1 = myDist;
        int fruits2 = other.myDist;
        int cmp = fruits1 - fruits2;
        for (int i = 1; i < turnsToGetExtraFruits.length; ++i) {
            int nextFruits1 = turnsToGetExtraFruits[i];
            int nextFruits2 = other.turnsToGetExtraFruits[i];
            if (nextFruits1 == Integer.MAX_VALUE && nextFruits2 == Integer.MAX_VALUE) {
                break;
            }
            fruits1 = calcFruits(fruits1, nextFruits1);
            fruits2 = calcFruits(fruits2, nextFruits2);
            if (cmp == 0) {
                cmp = fruits1 - fruits2;
            } else if (cmp < 0 && fruits2 < fruits1) {
                cmp = 1;
            } else if (cmp > 0 && fruits1 < fruits2) {
                cmp = -1;
            }
        }
        return cmp;
    }

    public void addCloseFruits(Map<Position, Map<Position, Integer>> fruitToFruitDistance) {
        turnsToGetExtraFruits[0] = myDist;
        for (int i = 1; i <= MAX_CLOSE_EXTRA_FRUITS; ++i) {
            turnsToGetExtraFruits[i] = Integer.MAX_VALUE;
        }
        Set<Position> seen = new HashSet<Position>();
        calc(fruitPosition, MAX_CLOSE_EXTRA_FRUITS, myDist, seen, fruitToFruitDistance);
    }

    private int calcFruits(int fruits, int nextFruits) {
        if (nextFruits == Integer.MAX_VALUE) {
            return fruits + AVERAGE_TURNS_TO_NEXT_FRUIT;
        }
        return Math.min(fruits + AVERAGE_TURNS_TO_NEXT_FRUIT, nextFruits);
    }

    private int getMaxTotalFruitDist() {
        return Math.min(MAX_TOTAL_FRUIT_DIST, TURNS_PER_GAME - turn);
    }

    private void calc(Position currentPos, int left, int totDist, Set<Position> seen,
                      Map<Position, Map<Position, Integer>> fruitToFruitDistance) {
        if (left == 0) {
            return;
        }
        --left;
        Map<Position, Integer> fruitDistance = fruitToFruitDistance.get(currentPos);
        if (fruitDistance != null) {
            for (Map.Entry<Position, Integer> entry : fruitDistance.entrySet()) {
                int newDist = totDist + entry.getValue();
                if (newDist <= getMaxTotalFruitDist() && !seen.contains(entry.getKey())) {
                    seen.add(entry.getKey());
                    if (newDist < turnsToGetExtraFruits[MAX_CLOSE_EXTRA_FRUITS - left]) {
                        turnsToGetExtraFruits[MAX_CLOSE_EXTRA_FRUITS - left] = newDist;
                    }
                    calc(entry.getKey(), left, newDist, seen, fruitToFruitDistance);
                    seen.remove(entry.getKey());
                }
            }
        }
    }
}
