package se.paxport.yggejonssoncrazysnake;

import se.citerus.crazysnake.BaseBrain;
import se.citerus.crazysnake.BrainId;
import se.citerus.crazysnake.Direction;
import se.citerus.crazysnake.HeatMeta;
import se.citerus.crazysnake.HeatState;
import se.citerus.crazysnake.Movement;
import se.citerus.crazysnake.Position;
import se.citerus.crazysnake.Snake;
import se.citerus.crazysnake.SquareContent;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Set;

@SuppressWarnings({"UnusedDeclaration"})
public class Paxport extends BaseBrain {

    protected static final int LETHAL = -(1 << 28);

    private static final int FRUIT = -(1 << 27);

    private static final int MAX_LOOK_AHEAD_TURNS_FOR_BLOCKING = 6;

    private HeatState state;

    private int maxX;

    private int maxY;

    private Position headPosition;

    @SuppressWarnings({"FieldCanBeLocal"})
    private Direction currentDirection;

    private final Direction[] dirValues = Direction.values();

    private final Movement[] movValues = Movement.values();

    @SuppressWarnings({"FieldCanBeLocal"})
    private Snake me;

    @SuppressWarnings({"FieldCanBeLocal"})
    private Snake opponent;

    /**
     * A view of the snakes on the game arena that contains the structure of each snake,
     * unlike the official {@link Snake} representation.
     */
    private final Map<BrainId, SnakeWithSegments> snakesWithSegments = new HashMap<BrainId, SnakeWithSegments>();

    /**
     * Occupied status of cells on the game arena. Can be either 0 (unoccupied), {@link #FRUIT}, {@link #LETHAL},
     * or the number of turns until the cell becomes empty (if the cell is currently occupied by a snake).
     */
    private int[][] occupied = null;

    /**
     * Gå inte direkt mot den här, använd {@link #getZeroMatrix()} istället.
     */
    private int[][] occupiedTmp = null;

    private int[] zeroRow = null;

    @Override
    public void init(Set<BrainId> participants, HeatMeta meta) {
        super.init(participants, meta);
        snakesWithSegments.clear();
    }

    private static int manhattanDist(Position p1, Position p2) {
        return Math.abs(p1.getX() - p2.getX()) + Math.abs(p1.getY() - p2.getY());
    }

    @SuppressWarnings({"ConstantConditions"})
    @Override
    public Movement getNextMove(HeatState state) {
        this.state = state;
        maxY = state.getNoRows();
        maxX = state.getNoColumns();
        me = getMe(state);
        opponent = getOpponent(state);
        headPosition = me.getHeadPosition();
        currentDirection = me.getDirection();

        updateSnakesWithSegments(state);
        preCalcOccupied();

        List<Position> goalPos = blockOpponentIfPossible();

        if (goalPos.isEmpty()) {
            goalPos = calculateGoalPosFromFruits();
        }

        Map<Direction, Integer> priorityDirMap = new HashMap<Direction, Integer>();
        priorityDirMap.put(Direction.EAST, Integer.MAX_VALUE);
        priorityDirMap.put(Direction.SOUTH, Integer.MAX_VALUE);
        priorityDirMap.put(Direction.WEST, Integer.MAX_VALUE);
        priorityDirMap.put(Direction.NORTH, Integer.MAX_VALUE);
        List<Direction> priorityDir = new ArrayList<Direction>();
        priorityDir.add(Direction.EAST);
        priorityDir.add(Direction.SOUTH);
        priorityDir.add(Direction.WEST);
        priorityDir.add(Direction.NORTH);
        int realPriorityDir = 0;
        if (!goalPos.isEmpty()) {
            int[][] g = getZeroMatrix();
            Direction[][] from = new Direction[maxY][maxX];

            g[headPosition.getY()][headPosition.getX()] = 1;
            Queue<Position> q = new LinkedList<Position>();
            q.add(headPosition);

            int seen = 0;
            while (!q.isEmpty() && seen < goalPos.size()) {
                Position p = q.remove();
                for (Direction dir : dirValues) {
                    if (p.equals(headPosition) && negateDir(dir).equals(currentDirection)) {
                        continue;
                    }
                    Position p2 = p.offset(dir);
                    if (p2.getX() >= 0 && p2.getX() < maxX && p2.getY() >= 0 && p2.getY() < maxY
                            && occupied[p2.getY()][p2.getX()] != LETHAL && g[p2.getY()][p2.getX()] <= 0
                            && (occupied[p2.getY()][p2.getX()] == FRUIT
                            || g[p.getY()][p.getX()] >= occupied[p2.getY()][p2.getX()])) {
                        from[p2.getY()][p2.getX()] = dir;
                        boolean found = false;
                        for (int i = 0; i < goalPos.size(); ++i) {
                            if (goalPos.get(i).equals(p2)) {
                                Direction d = Direction.SOUTH;
                                Position c = p2;
                                while (!c.equals(headPosition)) {
                                    d = from[c.getY()][c.getX()];
                                    c = c.offset(negateDir(d));
                                }
                                if (priorityDirMap.get(d) > i) {
                                    priorityDirMap.put(d, i);
                                    priorityDir.remove(d);
                                    ++realPriorityDir;
                                    boolean added = false;
                                    for (int j = 0; j < priorityDir.size(); ++j) {
                                        if (priorityDirMap.get(priorityDir.get(j)) > i) {
                                            priorityDir.add(j, d);
                                            added = true;
                                            break;
                                        }
                                    }
                                    if (!added) {
                                        priorityDir.add(d);
                                    }
                                }
                                found = true;
                                ++seen;
                                break;
                            }
                        }
                        if (!found) {
                            q.add(p2);
                        }
                        g[p2.getY()][p2.getX()] = g[p.getY()][p.getX()] + 1;
                    }
                }
            }
        }

        // En sista koll så att vi inte går in i en vägg
        Set<Movement> forbjoden = new HashSet<Movement>();
        for (Movement meMov : movValues) {
            Position meGoTo = headPosition.offset(currentDirection.turn(meMov));
            if (isOutsideOrLethal(meGoTo)) {
                forbjoden.add(meMov);
            }
        }

        List<Pair<Integer, Direction>> secondPriority = new ArrayList<Pair<Integer, Direction>>();
        int openSpaceLimit = Math.min(me.getLength() + (me.getLength() / meta.getGrowFrequency()) + 17, 400);
        for (int i = 0; i < priorityDir.size(); ++i) {
            Direction dir = priorityDir.get(i);
            Movement movement = Direction.newMovement(currentDirection, dir);
            if (movement == null
                    || oppositeDirections(currentDirection, dir)
                    || isOutsideOrLethal(headPosition.offset(currentDirection.turn(movement)))
                    || forbjoden.contains(movement)) {
                priorityDir.remove(i);
                --i;
                continue;
            }
            int myOpenSpace = calcOpenSpace(dir, Integer.MAX_VALUE);
            if (myOpenSpace < openSpaceLimit) {
                secondPriority.add(new Pair<Integer, Direction>(myOpenSpace, dir));
                priorityDir.remove(i);
                --i;
            } else if (priorityDirMap.get(dir) == Integer.MAX_VALUE) {
                secondPriority.add(new Pair<Integer, Direction>(myOpenSpace, dir));
                priorityDir.remove(i);
                --i;
            }
        }
        if (priorityDir.isEmpty()) {
            if (secondPriority.isEmpty()) {
                Direction d = directionTowardsCenterOfBoard();
                if (priorityDir.contains(d)) {
                    priorityDir.remove(d);
                }
                priorityDir.add(0, d);
            } else {
                Collections.sort(secondPriority);
                Direction dir = secondPriority.get(secondPriority.size() - 1).getSecond();
                int p = secondPriority.get(secondPriority.size() - 1).getFirst();
                for (int i = secondPriority.size() - 2; i >= 0; --i) {
                    Pair<Integer, Direction> prio = secondPriority.get(i);
                    if (prio.getFirst() != p) {
                        break;
                    }
                    if (!towardsCenter(dir) && towardsCenter(prio.getSecond())) {
                        dir = prio.getSecond();
                    }
                }
                priorityDir.add(dir);
            }
        }
        Movement mov;
        if (priorityDir.isEmpty()) {
            mov = Movement.FORWARD;
        } else {
            mov = Direction.newMovement(currentDirection, priorityDir.get(0));
        }
        Position meNextPos = me.getHeadPosition().offset(currentDirection.turn(mov));
        if (occupied[meNextPos.getY()][meNextPos.getX()] == 1) {
            System.out.println("Nära-svansen-upplevelse");
        }
        return mov;
    }

    private boolean towardsCenter(Direction direction) {
        int dx = me.getHeadPosition().getX() - state.getNoColumns() / 2;
        int dy = me.getHeadPosition().getY() - state.getNoRows() / 2;
        if (direction == Direction.NORTH) {
            return dy > 0;
        } else if (direction == Direction.WEST) {
            return dx > 0;
        } else if (direction == Direction.SOUTH) {
            return dy < 0;
        } else {
            return dx < 0;
        }
    }

    private Direction directionTowardsCenterOfBoard() {
        int dx = me.getHeadPosition().getX() - state.getNoColumns() / 2;
        int dy = me.getHeadPosition().getY() - state.getNoRows() / 2;
        if (Math.abs(dx) > Math.abs(dy)) {
            Direction wantedDirection = dx < 0 ? Direction.EAST : Direction.WEST;
            if (oppositeDirections(currentDirection, wantedDirection)) {
                return dy < 0 ? Direction.SOUTH : Direction.NORTH;
            }
            return wantedDirection;
        } else {
            Direction wantedDirection = dy < 0 ? Direction.SOUTH : Direction.NORTH;
            if (oppositeDirections(currentDirection, wantedDirection)) {
                return dx < 0 ? Direction.EAST : Direction.WEST;
            }
            return wantedDirection;
        }
    }

    private boolean oppositeDirections(Direction firstDirection, Direction secondDirection) {
        if (firstDirection == Direction.NORTH) {
            return secondDirection == Direction.SOUTH;
        } else if (firstDirection == Direction.EAST) {
            return secondDirection == Direction.WEST;
        } else if (firstDirection == Direction.SOUTH) {
            return secondDirection == Direction.NORTH;
        } else {
            return secondDirection == Direction.EAST;
        }
    }

    private List<Position> blockOpponentIfPossible() {
        int requiredOpenSpace = opponent.getLength();
        int currentOpponentOpenSpace = calculateOpenSpace(opponent.getHeadPosition(), 0, requiredOpenSpace);
        if (currentOpponentOpenSpace < requiredOpenSpace) {
            System.out.println("Opponent redan dödsdömd");
            return new ArrayList<Position>();
        }

        for (int turnsToCheck = 1; turnsToCheck <= MAX_LOOK_AHEAD_TURNS_FOR_BLOCKING; ++turnsToCheck) {
            Position pos = calcMove(headPosition, currentDirection,
                    opponent.getHeadPosition(), opponent.getDirection(),
                    turnsToCheck);
            if (pos != null) {
                System.out.println("Kill!");
                return Arrays.asList(pos);
            }
        }
        return new ArrayList<Position>();
    }

    private Position calcMove(Position myPosition, Direction myDirection,
                              Position opponentPosition, Direction opponentDirection,
                              int turnsForBlocking) {
        int requiredOpenSpace = opponent.getLength();
        for (Movement myMov : movValues) {
            Position myPos = myPosition.offset(myDirection.turn(myMov));
            if (isOutsideOrLethal(myPos)) {
                continue;
            }

            boolean foundOpenSpace = false;
            int myPrev = occupied[myPos.getY()][myPos.getX()];
            occupied[myPos.getY()][myPos.getX()] = LETHAL;
            for (Movement oppMove : movValues) {
                Position oppPos = opponentPosition.offset(opponentDirection.turn(oppMove));
                if (oppPos.equals(myPos)) {
                    foundOpenSpace = true;
                    break;
                }
                if (isOutsideOrLethal(oppPos)) {
                    continue;
                }
                if (turnsForBlocking == 1) {
                    int openSpace = calculateOpenSpace(oppPos, 1, requiredOpenSpace);
                    if (openSpace >= requiredOpenSpace) {
                        foundOpenSpace = true;
                        break;
                    }
                } else {
                    Position posIfBlockSuccessful =
                            calcMove(myPos, myDirection.turn(myMov),
                                    oppPos, opponentDirection.turn(oppMove),
                                    turnsForBlocking - 1);
                    if (posIfBlockSuccessful == null) {
                        foundOpenSpace = true;
                        break;
                    }
                }
            }
            occupied[myPos.getY()][myPos.getX()] = myPrev;
            if (!foundOpenSpace) {
                return myPos;
            }
        }
        return null;
    }

    private List<Position> calculateGoalPosFromFruits() {
        // Vem är närmast varje frukt och hur långt har de till den?
        Map<Position, Map.Entry<Snake, Integer>> nearestToFruit = new HashMap<Position, Map.Entry<Snake, Integer>>();
        Map<Position, Integer> myDists = calcDistanceToFruits(me);
        Map<Position, Integer> opponentDists = calcDistanceToFruits(opponent);
        Map<Position, Map<Position, Integer>> fruitToFruitDistance = new HashMap<Position, Map<Position, Integer>>();
        FruitDistanceSorter sorter = new FruitDistanceSorter();
        List<FruitDistance> oppDistances = new ArrayList<FruitDistance>();
        for (Map.Entry<Position, Integer> myFruitDist : myDists.entrySet()) {
            Position fruitPos = myFruitDist.getKey();
            int myDist = myFruitDist.getValue();
            Integer oppDist = opponentDists.get(fruitPos);
            FruitDistance distance = new FruitDistance((int) state.getTurn(), fruitPos, myDist, oppDist);
            if (oppDist == null || myDist <= oppDist) {
                Map<Position, Integer> fruitDistances = calcDistanceToFruits(fruitPos, null, myDist);
                fruitToFruitDistance.put(fruitPos, fruitDistances);
                fruitDistances.remove(fruitPos);
                sorter.add(distance);
            } else {
                oppDistances.add(distance);
            }
        }
        sorter.addCloseFruits(fruitToFruitDistance);
        List<Position> distances = sorter.getDistances();
        if (distances.isEmpty()) {
            Collections.sort(oppDistances, new Comparator<FruitDistance>() {
                @Override
                public int compare(FruitDistance o1, FruitDistance o2) {
                    return o2.getOppDist() - o1.getOppDist();
                }
            });
            for (FruitDistance oppDistance : oppDistances) {
                distances.add(oppDistance.getFruitPosition());
            }
        }
        return distances;
    }

    private Map<Position, Integer> calcDistanceToFruits(Snake snake) {
        return calcDistanceToFruits(snake.getHeadPosition(), snake.getDirection(), 1);
    }

    private Map<Position, Integer> calcDistanceToFruits(Position startPos, Direction startDir, int startCount) {
        int[][] g = getZeroMatrix();

        g[startPos.getY()][startPos.getX()] = startCount;
        Queue<Position> q = new LinkedList<Position>();
        q.add(startPos);

        Map<Position, Integer> fruitDistances = new HashMap<Position, Integer>();

        while (!q.isEmpty()) {
            Position p = q.remove();
            for (Direction dir : dirValues) {
                if (p.equals(startPos) && negateDir(dir).equals(startDir)) {
                    continue;
                }
                Position p2 = p.offset(dir);
                if (p2.getX() >= 0 && p2.getX() < maxX && p2.getY() >= 0 && p2.getY() < maxY
                        && occupied[p2.getY()][p2.getX()] != LETHAL && g[p2.getY()][p2.getX()] <= 0
                        && (occupied[p2.getY()][p2.getX()] == FRUIT
                        || g[p.getY()][p.getX()] >= occupied[p2.getY()][p2.getX()])) {
                    q.add(p2);
                    if (occupied[p2.getY()][p2.getX()] == FRUIT) {
                        fruitDistances.put(p2, g[p.getY()][p.getX()]);
                    }
                    g[p2.getY()][p2.getX()] = g[p.getY()][p.getX()] + 1;
                }
            }
        }

        return fruitDistances;
    }

    void updateSnakesWithSegments(HeatState state) {
        if (snakesWithSegments.isEmpty()) {
            for (Snake snake : state.snakes()) {
                snakesWithSegments.put(snake.getId(), new SnakeWithSegments(snake));
            }
        }
        for (Snake snake : state.snakes()) {
            snakesWithSegments.get(snake.getId()).refreshState(state);
        }
    }

    /**
     * Endast en användare av resultatet åt gången tack!
     *
     * @return nollad återanvänd matris
     */
    private int[][] getZeroMatrix() {
        if (occupiedTmp == null) {
            occupiedTmp = new int[maxY][maxX];
            zeroRow = new int[maxX];
        }
        for (int y = 0; y < maxY; y++) {
            System.arraycopy(zeroRow, 0, occupiedTmp[y], 0, maxX);
        }
        return occupiedTmp;
    }

    private void preCalcOccupied() {
        if (occupied == null) {
            occupied = new int[maxY][maxX];
        }

        for (int y = 0; y < maxY; y++) {
            for (int x = 0; x < maxX; x++) {
                if (isLethalCollision(x, y)) {
                    occupied[y][x] = LETHAL;
                } else if (state.getSquare(x, y).containsFruit()) {
                    occupied[y][x] = FRUIT;
                } else {
                    occupied[y][x] = 0;
                }
            }
        }

        for (SnakeWithSegments snakeWithSegments : snakesWithSegments.values()) {
            snakeWithSegments.populateSnakeShrinking(occupied, meta, state);
        }
    }

    private int calcOpenSpace(Direction dir2, int maxLimit) {
        Position newHead = headPosition.offset(dir2);

        if (isOutsideOrLethal(newHead)) {
            return -1000;
        }

        return calculateOpenSpace(newHead, 0, maxLimit);
    }

    private int calculateOpenSpace(Position newHead, int turnsPlayedBeforeCalc, int maxLimit) {
        int[][] g = getZeroMatrix();
        g[newHead.getY()][newHead.getX()] = 1 + turnsPlayedBeforeCalc;
        Queue<Position> q = new LinkedList<Position>();
        q.add(newHead);

        int openSpace = 0;
        while (!q.isEmpty()) {
            openSpace++;
            if (openSpace > maxLimit) {
                return openSpace;
            }
            Position p = q.remove();
            for (Direction dir : dirValues) {
                Position p2 = p.offset(dir);
                if (p2.getX() >= 0 && p2.getX() < maxX && p2.getY() >= 0 && p2.getY() < maxY
                        && occupied[p2.getY()][p2.getX()] != LETHAL && g[p2.getY()][p2.getX()] == 0
                        && (occupied[p2.getY()][p2.getX()] == FRUIT
                        || g[p.getY()][p.getX()] >= occupied[p2.getY()][p2.getX()])) {
                    q.add(p2);
                    g[p2.getY()][p2.getX()] = g[p.getY()][p.getX()] + 1;
                }
            }
        }
        return openSpace;
    }

    private boolean isOutsideOrLethal(Position position) {
        final int x = position.getX(), y = position.getY();
        return x < 0 || x >= maxX || y < 0 || y >= maxY
                || occupied[y][x] == LETHAL
                || occupied[y][x] > 1;
    }

    private static Direction negateDir(Direction dir) {
        switch (dir) {
            case SOUTH:
                return Direction.NORTH;
            case NORTH:
                return Direction.SOUTH;
            case WEST:
                return Direction.EAST;
            case EAST:
                return Direction.WEST;
        }
        return null;
    }

    private boolean isLethalCollision(Position p) {
        return isLethalCollision(p.getX(), p.getY());
    }

    private boolean isLethalCollision(int x, int y) {
        if (x < 0 || x >= maxX || y < 0 || y >= maxY) {
            return true;
        }
        SquareContent content = state.getSquare(x, y).getContent();
        return content.getContentType().isSolid();
    }
}
