package se.paxport.yggejonssoncrazysnake;

import se.citerus.crazysnake.*;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

/**
 * @author Joel Jonsson
 */
public class SnakeWithSegments {

    private final LinkedList<Position> segments = new LinkedList<Position>();

    private final BrainId id;

    private List<Integer> tmpOnExtraVag = new ArrayList<Integer>();
    private int turnsUntilGrowth;
    private int turnsUntilGone;
    private int turnsUntilGrowthOnTurn;
    private int turnsUntilGoneOnTurn;

    public SnakeWithSegments(Snake snake) {
        this.id = snake.getId();
        segments.add(snake.getHeadPosition());
        if (snake.getLength() != 1) {
            System.out.println(String.format("Snake %s has unexpected length: %d", snake.getName(), snake.getLength()));
        }
    }

    public void refreshState(HeatState state) {
        shrinkTail(state);
        growHead(state);
        if (segments.size() != state.getSnake(id).getLength()) {
            System.out.println("SnakeWithSegments.refreshState length wrong");
            System.out.println("segments.size() = " + segments.size());
            System.out.println("state.getSnake(id).getLength() = " + state.getSnake(id).getLength());
        }
    }

    private void growHead(HeatState heatState) {
        Position headPosition = heatState.getSnake(id).getHeadPosition();
        if (segments.isEmpty() || !segments.getFirst().equals(headPosition)) {
            segments.addFirst(headPosition);
        }
    }

    private void shrinkTail(HeatState heatState) {
        while (!segments.isEmpty()) {
            Position tail = segments.getLast();
            Square square = heatState.getSquare(tail);
            if (!square.containsSnake()
                    || square.getContent().getContentType() == SquareContentType.SNAKE_HEAD
                    || !id.equals(square.getContent().getSnake().getId())) {
                segments.removeLast();
            } else {
                break;
            }
        }
    }

    public void populateSnakeShrinking(int[][] g, HeatMeta meta, HeatState state) {
        boolean earlyTurn = state.getTurn() <= 10;
        turnsUntilGrowth = (int) state.getTurnsUntilGrowth();
        turnsUntilGone = 0;
        Iterator<Position> tailToHead = segments.descendingIterator();
        while (tailToHead.hasNext()) {
            turnsUntilGone++;
            if (turnsUntilGrowth == 0) {
                turnsUntilGone++;
                turnsUntilGrowth = meta.getGrowFrequency();
            } else {
                turnsUntilGrowth--;
            }
            Position pos = tailToHead.next();
            // Det är ok att gå till pos så att du är där i turnsUntilGone drag
            g[pos.getY()][pos.getX()] = earlyTurn ? Paxport.LETHAL : turnsUntilGone;
        }
        turnsUntilGrowthOnTurn = turnsUntilGone;
        turnsUntilGoneOnTurn = turnsUntilGone;
    }

    public void populateSnakeShrinkingWithExtraVag(int[][] g, HeatMeta meta, HeatState state, Position extraVag) {
        boolean earlyTurn = state.getTurn() <= 10;
        turnsUntilGone++;
        if (turnsUntilGrowth == 0) {
            turnsUntilGone++;
            turnsUntilGrowth = meta.getGrowFrequency();
        } else {
            turnsUntilGrowth--;
        }
        tmpOnExtraVag.add(g[extraVag.getY()][extraVag.getX()]);
        // Det är ok att gå till pos så att du är där i turnsUntilGone drag
        g[extraVag.getY()][extraVag.getX()] = earlyTurn ? Paxport.LETHAL : turnsUntilGone;
    }

    public void clearExtraVag(int[][] g, List<Position> extraVag) {
        if(extraVag.size() != tmpOnExtraVag.size())
            System.out.println("FEEEEEL");
        Iterator<Integer> extraIt = tmpOnExtraVag.iterator();
        for (Position pos : extraVag) {
            g[pos.getY()][pos.getX()] = extraIt.next();
        }
        tmpOnExtraVag.clear();
        turnsUntilGone = turnsUntilGrowthOnTurn;
        turnsUntilGone = turnsUntilGoneOnTurn;
    }
}
