package se.paxport.yggejonssoncrazysnake;

import org.junit.Before;
import org.junit.Test;
import se.citerus.crazysnake.Direction;
import se.citerus.crazysnake.Movement;
import se.citerus.crazysnake.Position;

public class Paxport_RaceCloseToWall_UnitTest extends AbstractPaxportTest {

    private static final String[] BOARD = new String[]{
            "#############",
            "#...........#",
            "#..@........#",
            "#...........#",
            "#...........#",
            "#...........#",
            "#############",
    };

    @Before
    public void setUp() throws Exception {
        initBoard(BOARD);
    }

    @Test
    public void givenOurSnakeIsBehindThenGotoNearestFruit() {
        setupSnakes(Direction.EAST, new Position(7, 4), new Position(8, 5));

        assertMovement(Movement.LEFT);
    }

    @Test
    public void givenOurSnakeIsAheadThenBlockOpponent() {
        setupSnakes(Direction.EAST, new Position(9, 4), new Position(7, 5));

        assertMovement(Movement.RIGHT);
    }

    @Test
    public void givenOurSnakeIsAheadAndCloseToWallThenContinueToBlockOpponent() {
        setupSnakes(Direction.EAST, new Position(10, 4), new Position(7, 5));

        assertMovement(Movement.FORWARD);
    }

    @Test
    public void givenOurSnakeIsSlightlyAheadThenBlockLaterIfThatWouldCauseCrashThisTurnAndWeHaveEqualScore() {
        setupSnakes(Direction.EAST, new Position(8, 4), new Position(7, 5), 15, 15);

        assertMovement(Movement.FORWARD);
    }

    @Test
    public void givenOurSnakeIsSlightlyAheadThenBlockLaterIfThatWouldCauseCrashThisTurnAndWeHaveMoreScore() {
        setupSnakes(Direction.EAST, new Position(8, 4), new Position(7, 5), 15, 14);

        assertMovement(Movement.FORWARD);
    }

    @Test
    public void givenOurSnakeIsSlightlyAheadThenBlockLaterIfThatWouldCauseCrashThisTurnAndWeHaveLessScore() {
        setupSnakes(Direction.EAST, new Position(8, 4), new Position(7, 5), 14, 15);

        assertMovement(Movement.FORWARD);
    }

    /**
     * Same tests as above, but mirrored to make sure that direction does not matter
     */

    @Test
    public void givenOurSnakeIsBehindMirrorThenGotoNearestFruit() {
        setupSnakes(Direction.WEST, new Position(5, 4), new Position(4, 5));

        assertMovement(Movement.RIGHT);
    }

    @Test
    public void givenOurSnakeIsAheadMirrorThenBlockOpponent() {
        setupSnakes(Direction.WEST, new Position(3, 4), new Position(5, 5));

        assertMovement(Movement.LEFT);
    }

    @Test
    public void givenOurSnakeIsAheadAndCloseToWallMirrorThenContinueToBlockOpponent() {
        setupSnakes(Direction.WEST, new Position(2, 4), new Position(5, 5));

        assertMovement(Movement.FORWARD);
    }

    @Test
    public void givenOurSnakeIsSlightlyAheadMirrorThenBlockLaterIfThatWouldCauseCrashThisTurnAndWeHaveMoreScore() {
        setupSnakes(Direction.WEST, new Position(3, 4), new Position(4, 5), 15, 14);

        assertMovement(Movement.FORWARD);
    }

    @Test
    public void givenOurSnakeIsSlightlyAheadMirrorThenBlockLaterIfThatWouldCauseCrashThisTurnAndWeHaveEqualScore() {
        setupSnakes(Direction.WEST, new Position(3, 4), new Position(4, 5), 15, 15);

        assertMovement(Movement.FORWARD);
    }

    @Test
    public void givenOurSnakeIsSlightlyAheadMirrorThenBlockLaterIfThatWouldCauseCrashThisTurnAndWeHaveLessScore() {
        setupSnakes(Direction.WEST, new Position(3, 4), new Position(4, 5), 14, 15);

        assertMovement(Movement.FORWARD);
    }

    private void setupSnakes(Direction direction, Position myPos, Position oppPos) {
        setupSnakes(new SnakeSetup(direction, myPos, 5),
                new SnakeSetup(direction, oppPos, 5),
                true);

    }

    private void setupSnakes(Direction direction, Position myPos, Position oppPos, int myScore, int oppScore) {
        setupSnakes(new SnakeSetup(direction, myPos, 5, myScore),
                new SnakeSetup(direction, oppPos, 5, oppScore),
                true);
    }
}
