package se.paxport.yggejonssoncrazysnake.util;

import se.citerus.crazysnake.*;

import java.awt.*;

public class Board {

    private static final SquareContentType FRUIT = SquareContentType.APPLE;
    private final String[] board;
    private final Snake currentSnake, opponentSnake;

    public Board(String[] board, Snake currentSnake, Snake opponentSnake) {
        this.board = new String[board.length];
        this.currentSnake = currentSnake;
        this.opponentSnake = opponentSnake;
        System.arraycopy(board, 0, this.board, 0, this.board.length);
    }

    public Square[][] getSquares() {
        final Square[][] squares = new Square[board[0].length()][];
        for (int x = 0; x < squares.length; ++x) {
            squares[x] = new Square[board.length];
            for (int y = 0; y < board.length; ++y) {
                squares[x][y] = new BoardSquare(x, y);
            }
        }
        return squares;
    }

    public Dimension getDimension() {
        return new Dimension(board[0].length(), board.length);
    }

    public void updateSnakePositions() {
        updateSnakePositions(currentSnake, 'a');
        updateSnakePositions(opponentSnake, 'A');
        for (String b : board) {
            System.out.println(b);
        }
    }

    public String[] getBoard() {
        return board;
    }

    private SquareContentType getSquareContentType(int x, int y) {
        char c = board[y].charAt(x);
        if (c == '.') {
            return SquareContentType.EMPTY;
        } else if (c == '#') {
            return SquareContentType.WALL;
        } else if (c == '@') {
            return SquareContentType.APPLE;
        } else if (c == 'a' || c == 'A') {
            return SquareContentType.SNAKE_HEAD;
        } else if ((c >= 'b' && c <= 'z') || (c >= 'B' && c <= 'Z')) {
            return SquareContentType.SNAKE_TAIL;
        } else {
            throw new RuntimeException(String.format("Invalid char %c on board", c));
        }
    }

    private Snake getSnakeOnPos(int x, int y) {
        char c = board[y].charAt(x);
        return c >= 'a' && c <= 'z' ? currentSnake : opponentSnake;
    }

    private void updateSnakePositions(Snake snake, char c) {
        int x = snake.getHeadPosition().getX();
        int y = snake.getHeadPosition().getY();
        setChar(c, y, x);
        int dx = 0;
        int dy = 0;
        if (snake.getDirection() == Direction.EAST) {
            --dx;
        } else if (snake.getDirection() == Direction.WEST) {
            ++dx;
        } else if (snake.getDirection() == Direction.NORTH) {
            --dy;
        } else if (snake.getDirection() == Direction.SOUTH) {
            ++dy;
        }
        for (int i = 1; i < snake.getLength(); ++i) {
            setChar(++c, y += dy, x += dx);
        }
    }

    private void setChar(char c, int y, int x) {
        board[y] = board[y].substring(0, x) + c + board[y].substring(x + 1);
    }

    private final class BoardSquare implements Square {

        private final int x, y;

        public BoardSquare(int x, int y) {
            this.x = x;
            this.y = y;
        }

        public boolean isUnoccupied() {
            return getSquareContentType(x, y) == SquareContentType.EMPTY;
        }

        public SquareContent getContent() {
            return new SquareContent() {
                public SquareContentType getContentType() {
                    return getSquareContentType(x, y);
                }

                public Snake getSnake() {
                    return getSnakeOnPos(x, y);
                }
            };
        }

        public boolean containsSnake() {
            return getSquareContentType(x, y) == SquareContentType.SNAKE_HEAD
                    || getSquareContentType(x, y) == SquareContentType.SNAKE_TAIL;
        }

        public boolean containsFruit() {
            return getSquareContentType(x, y) == FRUIT;
        }

        public boolean containsWall() {
            return getSquareContentType(x, y) == SquareContentType.WALL;
        }

        public boolean hasSolidContent() {
            return containsSnake() || containsWall();
        }
    }
}
